@@ -1,6 +1,7 @@
 /*
  * Copyright 2005 Eric Anholt
  * Copyright 2005 Benjamin Herrenschmidt
+ * Copyright 2006 Tungsten Graphics, Inc.
  * All Rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -26,6 +27,7 @@
  *    Eric Anholt <anholt@FreeBSD.org>
  *    Zack Rusin <zrusin@trolltech.com>
  *    Benjamin Herrenschmidt <benh@kernel.crashing.org>
+ *    Michel Dänzer <michel@tungstengraphics.com>
  *
  */
 
@@ -49,6 +51,7 @@
 #endif
 #endif
 
+#include <errno.h>
 #include <string.h>
 
 #include "radeon.h"
@@ -56,14 +59,30 @@
 
 #include "exa.h"
 
-#include "fbdevhw.h"
+static int
+FUNC_NAME(RADEONMarkSync)(ScreenPtr pScreen)
+{
+    RINFO_FROM_SCREEN(pScreen);
+
+    TRACE;
+
+    return ++info->exaSyncMarker;
+}
 
 static void
 FUNC_NAME(RADEONSync)(ScreenPtr pScreen, int marker)
 {
+    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
+    RADEONInfoPtr info = RADEONPTR(pScrn);
+
     TRACE;
 
-    FUNC_NAME(RADEONWaitForIdle)(xf86Screens[pScreen->myNum]);
+    if (info->exaMarkerSynced != marker) {
+	FUNC_NAME(RADEONWaitForIdle)(pScrn);
+	info->exaMarkerSynced = marker;
+    }
+
+    RADEONPTR(pScrn)->engineMode = EXA_ENGINEMODE_UNKNOWN;
 }
 
 static Bool
@@ -124,30 +143,14 @@ FUNC_NAME(RADEONDoneSolid)(PixmapPtr pPix)
     TRACE;
 }
 
-static Bool
-FUNC_NAME(RADEONPrepareCopy)(PixmapPtr pSrc,   PixmapPtr pDst,
-			     int xdir, int ydir,
-			     int rop,
-			     Pixel planemask)
+void
+FUNC_NAME(RADEONDoPrepareCopy)(ScrnInfoPtr pScrn, CARD32 src_pitch_offset,
+			       CARD32 dst_pitch_offset, CARD32 datatype, int rop,
+			       Pixel planemask)
 {
-    RINFO_FROM_SCREEN(pDst->drawable.pScreen);
-    CARD32 datatype, src_pitch_offset, dst_pitch_offset;
+    RADEONInfoPtr info = RADEONPTR(pScrn);
     ACCEL_PREAMBLE();
 
-    TRACE;
-
-    info->xdir = xdir;
-    info->ydir = ydir;
-
-    if (pDst->drawable.bitsPerPixel == 24)
-	RADEON_FALLBACK(("24bpp unsupported"));
-    if (!RADEONGetDatatypeBpp(pDst->drawable.bitsPerPixel, &datatype))
-	RADEON_FALLBACK(("RADEONGetDatatypeBpp failed\n"));
-    if (!RADEONGetPixmapOffsetPitch(pSrc, &src_pitch_offset))
-	RADEON_FALLBACK(("RADEONGetPixmapOffsetPitch source failed\n"));
-    if (!RADEONGetPixmapOffsetPitch(pDst, &dst_pitch_offset))
-	RADEON_FALLBACK(("RADEONGetPixmapOffsetPitch dest failed\n"));
-
     RADEON_SWITCH_TO_2D();
 
     BEGIN_ACCEL(5);
@@ -162,16 +165,43 @@ FUNC_NAME(RADEONPrepareCopy)(PixmapPtr pSrc,   PixmapPtr pDst,
 	RADEON_GMC_CLR_CMP_CNTL_DIS);
     OUT_ACCEL_REG(RADEON_DP_WRITE_MASK, planemask);
     OUT_ACCEL_REG(RADEON_DP_CNTL,
-	((xdir >= 0 ? RADEON_DST_X_LEFT_TO_RIGHT : 0) |
-	 (ydir >= 0 ? RADEON_DST_Y_TOP_TO_BOTTOM : 0)));
+	((info->xdir >= 0 ? RADEON_DST_X_LEFT_TO_RIGHT : 0) |
+	 (info->ydir >= 0 ? RADEON_DST_Y_TOP_TO_BOTTOM : 0)));
     OUT_ACCEL_REG(RADEON_DST_PITCH_OFFSET, dst_pitch_offset);
     OUT_ACCEL_REG(RADEON_SRC_PITCH_OFFSET, src_pitch_offset);
     FINISH_ACCEL();
+}
+
+static Bool
+FUNC_NAME(RADEONPrepareCopy)(PixmapPtr pSrc,   PixmapPtr pDst,
+			     int xdir, int ydir,
+			     int rop,
+			     Pixel planemask)
+{
+    RINFO_FROM_SCREEN(pDst->drawable.pScreen);
+    CARD32 datatype, src_pitch_offset, dst_pitch_offset;
+
+    TRACE;
+
+    info->xdir = xdir;
+    info->ydir = ydir;
+
+    if (pDst->drawable.bitsPerPixel == 24)
+	RADEON_FALLBACK(("24bpp unsupported"));
+    if (!RADEONGetDatatypeBpp(pDst->drawable.bitsPerPixel, &datatype))
+	RADEON_FALLBACK(("RADEONGetDatatypeBpp failed\n"));
+    if (!RADEONGetPixmapOffsetPitch(pSrc, &src_pitch_offset))
+	RADEON_FALLBACK(("RADEONGetPixmapOffsetPitch source failed\n"));
+    if (!RADEONGetPixmapOffsetPitch(pDst, &dst_pitch_offset))
+	RADEON_FALLBACK(("RADEONGetPixmapOffsetPitch dest failed\n"));
+
+    FUNC_NAME(RADEONDoPrepareCopy)(pScrn, src_pitch_offset, dst_pitch_offset,
+				   datatype, rop, planemask);
 
     return TRUE;
 }
 
-static void
+void
 FUNC_NAME(RADEONCopy)(PixmapPtr pDst,
 		      int srcX, int srcY,
 		      int dstX, int dstY,
@@ -211,10 +241,8 @@ static Bool
 FUNC_NAME(RADEONUploadToScreen)(PixmapPtr pDst, int x, int y, int w, int h,
 				char *src, int src_pitch)
 {
-#if X_BYTE_ORDER == X_BIG_ENDIAN || defined(ACCEL_CP)
     RINFO_FROM_SCREEN(pDst->drawable.pScreen);
-#endif
-    CARD8	   *dst	     = pDst->devPrivate.ptr;
+    CARD8	   *dst	     = info->FB + exaGetPixmapOffset(pDst);
     unsigned int   dst_pitch = exaGetPixmapPitch(pDst);
     unsigned int   bpp	     = pDst->drawable.bitsPerPixel;
 #ifdef ACCEL_CP
@@ -289,27 +317,152 @@ FUNC_NAME(RADEONUploadToScreen)(PixmapPtr pDst, int x, int y, int w, int h,
     return TRUE;
 }
 
+#ifdef ACCEL_CP
+/* Emit blit with arbitrary source and destination offsets and pitches */
+static void
+RADEONBlitChunk(ScrnInfoPtr pScrn, CARD32 datatype, CARD32 src_pitch_offset,
+		CARD32 dst_pitch_offset, int srcX, int srcY, int dstX, int dstY,
+		int w, int h)
+{
+    RADEONInfoPtr info = RADEONPTR(pScrn);
+    ACCEL_PREAMBLE();
+
+    BEGIN_ACCEL(6);
+    OUT_ACCEL_REG(RADEON_DP_GUI_MASTER_CNTL,
+		  RADEON_GMC_DST_PITCH_OFFSET_CNTL |
+		  RADEON_GMC_SRC_PITCH_OFFSET_CNTL |
+		  RADEON_GMC_BRUSH_NONE |
+		  (datatype << 8) |
+		  RADEON_GMC_SRC_DATATYPE_COLOR |
+		  RADEON_ROP3_S |
+		  RADEON_DP_SRC_SOURCE_MEMORY |
+		  RADEON_GMC_CLR_CMP_CNTL_DIS |
+		  RADEON_GMC_WR_MSK_DIS);
+    OUT_ACCEL_REG(RADEON_SRC_PITCH_OFFSET, src_pitch_offset);
+    OUT_ACCEL_REG(RADEON_DST_PITCH_OFFSET, dst_pitch_offset);
+    OUT_ACCEL_REG(RADEON_SRC_Y_X, (srcY << 16) | srcX);
+    OUT_ACCEL_REG(RADEON_DST_Y_X, (dstY << 16) | dstX);
+    OUT_ACCEL_REG(RADEON_DST_HEIGHT_WIDTH, (h << 16) | w);
+    FINISH_ACCEL();
+}
+#endif
+
 static Bool
 FUNC_NAME(RADEONDownloadFromScreen)(PixmapPtr pSrc, int x, int y, int w, int h,
 				    char *dst, int dst_pitch)
 {
-#if X_BYTE_ORDER == X_BIG_ENDIAN
     RINFO_FROM_SCREEN(pSrc->drawable.pScreen);
+#if X_BYTE_ORDER == X_BIG_ENDIAN
     unsigned char *RADEONMMIO = info->MMIO;
     unsigned int swapper = info->ModeReg.surface_cntl &
 	    ~(RADEON_NONSURF_AP0_SWP_32BPP | RADEON_NONSURF_AP1_SWP_32BPP |
 	      RADEON_NONSURF_AP0_SWP_16BPP | RADEON_NONSURF_AP1_SWP_16BPP);
 #endif
-    unsigned char *src	     = pSrc->devPrivate.ptr;
+    CARD8	  *src	     = info->FB + exaGetPixmapOffset(pSrc);
     int		   src_pitch = exaGetPixmapPitch(pSrc);
     int		   bpp	     = pSrc->drawable.bitsPerPixel;
+#ifdef ACCEL_CP
+    CARD32 datatype, src_pitch_offset, scratch_pitch = (w * bpp/8 + 63) & ~63, scratch_off = 0;
+    drmBufPtr scratch;
+#endif
 
     TRACE;
 
+#ifdef ACCEL_CP
     /*
-     * This is currently done without DMA until I have ironed out the
-     * various endian issues with R300 among others
+     * Try to accelerate download. Use an indirect buffer as scratch space,
+     * blitting the bits to one half while copying them out of the other one and
+     * then swapping the halves.
      */
+    if (info->accelDFS && bpp != 24 && RADEONGetDatatypeBpp(bpp, &datatype) &&
+	RADEONGetPixmapOffsetPitch(pSrc, &src_pitch_offset) &&
+	(scratch = RADEONCPGetBuffer(pScrn)))
+    {
+	int swap = RADEON_HOST_DATA_SWAP_NONE, wpass = w * bpp / 8;
+	int hpass = min(h, scratch->total/2 / scratch_pitch);
+	CARD32 scratch_pitch_offset = scratch_pitch << 16
+				    | (info->gartLocation + info->bufStart
+				       + scratch->idx * scratch->total) >> 10;
+	drmRadeonIndirect indirect;
+	ACCEL_PREAMBLE();
+
+	RADEON_SWITCH_TO_2D();
+
+	/* Kick the first blit as early as possible */
+	RADEONBlitChunk(pScrn, datatype, src_pitch_offset, scratch_pitch_offset,
+			x, y, 0, 0, w, hpass);
+	FLUSH_RING();
+
+#if X_BYTE_ORDER == X_BIG_ENDIAN
+	switch (bpp) {
+	case 16:
+	  swap = RADEON_HOST_DATA_SWAP_16BIT;
+	  break;
+	case 32:
+	  swap = RADEON_HOST_DATA_SWAP_32BIT;
+	  break;
+	}
+#endif
+
+	while (h) {
+	    int oldhpass = hpass, i = 0;
+
+	    src = (CARD8*)scratch->address + scratch_off;
+
+	    y += oldhpass;
+	    h -= oldhpass;
+	    hpass = min(h, scratch->total/2 / scratch_pitch);
+
+	    /* Prepare next blit if anything's left */
+	    if (hpass) {
+		scratch_off = scratch->total/2 - scratch_off;
+		RADEONBlitChunk(pScrn, datatype, src_pitch_offset, scratch_pitch_offset + (scratch_off >> 10),
+				x, y, 0, 0, w, hpass);
+	    }
+
+	    /*
+	     * Wait for previous blit to complete.
+	     *
+	     * XXX: Doing here essentially the same things this ioctl does in
+	     * the DRM results in corruption with 'small' transfers, apparently
+	     * because the data doesn't actually land in system RAM before the
+	     * memcpy. I suspect the ioctl helps mostly due to its latency; what
+	     * we'd really need is a way to reliably wait for the host interface
+	     * to be done with pushing the data to the host.
+	     */
+	    while ((drmCommandNone(info->drmFD, DRM_RADEON_CP_IDLE) == -EBUSY)
+		   && (i++ < RADEON_TIMEOUT))
+		;
+
+	    /* Kick next blit */
+	    if (hpass)
+		FLUSH_RING();
+
+	    /* Copy out data from previous blit */
+	    if (wpass == scratch_pitch && wpass == dst_pitch) {
+		RADEONCopySwap((CARD8*)dst, src, wpass * oldhpass, swap);
+		dst += dst_pitch * oldhpass;
+	    } else while (oldhpass--) {
+		RADEONCopySwap((CARD8*)dst, src, wpass, swap);
+		src += scratch_pitch;
+		dst += dst_pitch;
+	    }
+	}
+
+	indirect.idx = scratch->idx;
+	indirect.start = indirect.end = 0;
+	indirect.discard = 1;
+
+	drmCommandWriteRead(info->drmFD, DRM_RADEON_INDIRECT,
+			    &indirect, sizeof(drmRadeonIndirect));
+
+	info->exaMarkerSynced = info->exaSyncMarker;
+
+	return TRUE;
+    }
+#endif
+
+    /* Can't accelerate download */
     exaWaitSync(pSrc->drawable.pScreen);
 
 #if X_BYTE_ORDER == X_BIG_ENDIAN
@@ -354,8 +507,8 @@ Bool FUNC_NAME(RADEONDrawInit)(ScreenPtr pScreen)
 	return FALSE;
     }
 
-    info->exa->exa_major = 2;
-    info->exa->exa_minor = 0;
+    info->exa->exa_major = EXA_VERSION_MAJOR;
+    info->exa->exa_minor = EXA_VERSION_MINOR;
 
     info->exa->PrepareSolid = FUNC_NAME(RADEONPrepareSolid);
     info->exa->Solid = FUNC_NAME(RADEONSolid);
@@ -365,6 +518,7 @@ Bool FUNC_NAME(RADEONDrawInit)(ScreenPtr pScreen)
     info->exa->Copy = FUNC_NAME(RADEONCopy);
     info->exa->DoneCopy = FUNC_NAME(RADEONDoneCopy);
 
+    info->exa->MarkSync = FUNC_NAME(RADEONMarkSync);
     info->exa->WaitMarker = FUNC_NAME(RADEONSync);
     info->exa->UploadToScreen = FUNC_NAME(RADEONUploadToScreen);
     info->exa->DownloadFromScreen = FUNC_NAME(RADEONDownloadFromScreen);
@@ -378,9 +532,6 @@ Bool FUNC_NAME(RADEONDrawInit)(ScreenPtr pScreen)
     info->exa->pixmapOffsetAlign = RADEON_BUFFER_ALIGN + 1;
     info->exa->pixmapPitchAlign = 64;
 
-    info->exa->maxX = 2047;
-    info->exa->maxY = 2047;
-
 #ifdef RENDER
     if (info->RenderAccel) {
 	if (info->ChipFamily >= CHIP_FAMILY_R300) {
@@ -409,6 +560,16 @@ Bool FUNC_NAME(RADEONDrawInit)(ScreenPtr pScreen)
     }
 #endif
 
+#if EXA_VERSION_MAJOR > 2 || (EXA_VERSION_MAJOR == 2 && EXA_VERSION_MINOR >= 3)
+    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Setting EXA maxPitchBytes\n");
+
+    info->exa->maxPitchBytes = 16320;
+    info->exa->maxX = info->exa->Composite ? 2048 : 8192;
+#else
+    info->exa->maxX = info->exa->Composite ? 2048 : 16320 / 4;
+#endif
+    info->exa->maxY = info->exa->Composite ? 2048 : 8192;
+
     RADEONEngineInit(pScrn);
 
     if (!exaDriverInit(pScreen, info->exa)) {
